# -*- coding: utf-8 -*-
from __future__ import (
    division, absolute_import, print_function, unicode_literals,
)
from builtins import *                  # noqa
from future.builtins.disabled import *  # noqa

import re
from operator import methodcaller
from collections import defaultdict
import itertools

from zhlint.utils import (
    TextElement,
    count_newlines,
    count_offset,
    try_invoke
)


ZH_CHARACTERS = (
    r'[\u4e00-\u9fff]'
)

ZH_SYMBOLS = (
    r'['
    r'\u3000-\u303f'
    r'\uff00-\uff0f'
    r'\uff1a-\uff20'
    r'\uff3b-\uff40'
    r'\uff5b-\uff64'
    r'\uffe0-\uffee'
    r']'
)


# 1: a.
# 2: whitespaces.
# 3: b.
def single_space_patterns(
        a, b,
        a_join_b=True, b_join_a=True,
        a_non_preceding='', b_non_preceding=''):

    def join_non_preceding(non_preceding, text):
        if non_preceding:
            return r'(?:(?<!{0}))({1})'.format(non_preceding, text)
        else:
            return r'({0})'.format(text)

    # 1. no space.
    # prefix check.
    p11 = r'{a_component}(){b_component}'
    # suffix check.
    p12 = r'{b_component}(){a_component}'

    # 2. more than one whitespaces.
    # prefix check.
    p21 = (
        r'{a_component}'
        r'((?:(?!\n)\s){{2,}})'
        r'{b_component}'
    )
    # suffix check.
    p22 = (
        r'{b_component}'
        r'((?:(?!\n)\s){{2,}})'
        r'{a_component}'
    )

    # 3. wrong single whitespace: [\t\r\f\v]
    # only allow ' ' and '\n'.
    # prefix check.
    p31 = (
        r'{a_component}'
        r'((?:(?![ \n])\s){{1}})'
        r'{b_component}'
    )
    # suffix check.
    p32 = (
        r'{b_component}'
        r'((?:(?![ \n])\s){{1}})'
        r'{a_component}'
    )

    patterns = []
    if a_join_b:
        patterns.extend([
            p11,
            p21,
            p31,
        ])
    if b_join_a:
        patterns.extend([
            p12,
            p22,
            p32,
        ])

    return list(map(
        methodcaller(
            'format',
            a_component=join_non_preceding(a_non_preceding, a),
            b_component=join_non_preceding(b_non_preceding, b),
        ),
        patterns,
    ))


# 1: a.
# 2: whitespaces.
# 3: b.
def no_space_patterns(a, b):

    # prefix check.
    p1 = (
        r'({0})'
        r'((?:(?!\n)\s)+)'
        r'({1})'
    )
    # suffix check.
    p2 = (
        r'({1})'
        r'((?:(?!\n)\s)+)'
        r'({0})'
    )

    return list(map(
        methodcaller('format', a, b),
        [p1, p2],
    ))


def detect_by_patterns(patterns, element, ignore_matches=set()):

    for pattern in patterns:
        for m in re.finditer(pattern, element.content, re.UNICODE):
            if m.group(0) in ignore_matches:
                continue
            yield m


def detect_e101(element):

    return detect_by_patterns(
        single_space_patterns(ZH_CHARACTERS, r'[a-zA-z]'),
        element,
    )


def detect_e102(element):

    return detect_by_patterns(
        single_space_patterns(ZH_CHARACTERS, r'\d'),
        element,
    )


def detect_e103(element):

    p = (
        # non-digit, non-chinese
        r'(?:(?!\d|{0}|{1}|[!-/:-@\[-`{{-~]|\s))'
        # ignore ％, ℃, x, n.
        r'[^\uff05\u2103xn\%]'
    )
    p = p.format(ZH_CHARACTERS, ZH_SYMBOLS)

    return detect_by_patterns(
        single_space_patterns(
            r'\d+',
            p,
            b_join_a=False,
            a_non_preceding='\w',
        ),
        element,
    )


# 1: whitespaces.
# 2: left parenthesis.
# 3: digits.
# 4: right parenthesis.
def detect_e104(element):

    p1 = (
        r'(?<=[^\s]{1})'
        r'(\s*)'
        r'([(\uff08])'
        r'(\d+)'
        r'([)\uff09])'
    )

    # cover
    p2 = (
        r'(?:^|\n)'
        r'(\s*)'
        r'([(\uff08])'
        r'(\d+)'
        r'([)\uff09])'
    )

    patterns = [p1, p2]
    for i, p in enumerate(patterns):
        for m in detect_by_patterns([p], element):
            yield_tag = None
            if m.group(2) != '(' or m.group(4) != ')':
                yield_tag = 0
            # preceded by non-whitespace.
            if i == 0 and m.group(1) not in (' ', '\n'):
                yield_tag = 1
            # preceded by nothing or newline.
            if i == 1 and m.group(1) != '':
                yield_tag = 2

            if yield_tag is not None:
                yield m, yield_tag


def detect_e203(element):

    return detect_by_patterns(
        no_space_patterns(
            ZH_SYMBOLS,
            r'(?:(?!{0}|\s)).'.format(ZH_SYMBOLS),
        ),
        element,
    )


# 1: ellipsis.
def detect_e205(element):

    p = r'(\.{2,}|。{2,})'
    for m in re.finditer(p, element.content, flags=re.UNICODE):
        detected = m.group(0)
        if detected[0] != '.' or len(detected) != 6:
            yield m


# 1: duplicated marks.
def detect_e206(element):

    p1 = r'([!！]{2,})'
    p2 = r'([?？]{2,})'

    return detect_by_patterns(
        [p1, p2],
        element,
    )


# 1: ~
def detect_e207(element):

    p1 = r'(~+)'

    return detect_by_patterns(
        [p1],
        element,
    )


def contains_chinese_characters(content):
    return re.search(ZH_CHARACTERS, content, re.UNICODE)


def delimiter_in_email(content, m):
    OFFSET = 20
    i = m.start()
    lb = max(0, i - OFFSET)
    ub = i + OFFSET + 1

    segment = content[lb:ub]
    i = min(i, OFFSET)

    EMAIL = r"([a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+)"
    for email_m in re.finditer(EMAIL, segment):
        if email_m.start() <= i < email_m.end():
            return True
    return False


# simple url and filename.
def delimiter_in_simple_uri(content, m):
    i = m.start()

    if content[i] == '.':
        if i == 0 or i == len(content) - 1:
            return False
        else:
            return content[i - 1].isalnum() and content[i + 1].isalnum()

    if content[i] == ':':
        return i + 2 < len(content) and content[i:i + 3] == '://'

    return False


# 1: en punctuations.
# 2: whitespaces.
def detect_e201(element):
    if not contains_chinese_characters(element.content):
        return False

    PUNCTUATIONS = set('!"$\'(),.:;<>?[\\]^_{}')

    def marks_pattern(pattern_template, texts):

        patterns = []
        for text in texts:
            pattern = r'(?:{0})'.format(re.escape(text))
            patterns.append(pattern)

        return pattern_template.format(r'|'.join(patterns))

    def parenthesis_ranges(content):

        ranges = []
        for m in re.finditer(r'\(\d+\)', content, flags=re.UNICODE):
            ranges.append(
                (m.start(), m.end() - 1),
            )
        return ranges

    def match_in_ranges(match, ranges):
        mi = m.start()
        mj = m.end() - 1
        for ri, rj in ranges:
            if ri <= mi <= rj or ri <= mj <= rj:
                return True
        return False

    LATEX_MARKS = ['$$', '\(', '\)']

    ret = []

    # gerneral forms, ignore common shared characters: '@#%&+*-=|~'
    p1 = r'([{0}]+)(\s*)'.format(re.escape(''.join(PUNCTUATIONS)))

    # prefix
    p2 = ''.join(itertools.chain(
        marks_pattern(r'(?:(?<={0}))', LATEX_MARKS),
        p1,
    ))
    # suffix.
    p3 = ''.join(itertools.chain(
        p1,
        marks_pattern(r'(?:(?={0}))', LATEX_MARKS),
    ))
    # special cases: \(\).
    p4 = ''.join(itertools.chain(
        marks_pattern(r'(?:(?<={0}))', ['\(\)']),
        p1,
    ))
    # suffix.
    p5 = ''.join(itertools.chain(
        p1,
        marks_pattern(r'(?:(?={0}))', ['\(\)']),
    ))

    patterns = [p1, p2, p3, p4, p5]
    ranges = parenthesis_ranges(element.content)

    for m in detect_by_patterns(
        patterns,
        element,
        ignore_matches=set(['......']),
    ):
        if match_in_ranges(m, ranges):
            continue
        if SpecialWordHelper.delimiter_in_word(element.content, m):
            continue
        if delimiter_in_email(element.content, m):
            continue
        if delimiter_in_simple_uri(element.content, m):
            continue

        contains_mark = False
        for mark in LATEX_MARKS:
            if mark in m.group(0):
                contains_mark = True
                break
        if not contains_mark:
            ret.append(m)

    return ret


# 1: chineses.
def detect_e202(element):
    if contains_chinese_characters(element.content):
        return False

    return detect_by_patterns(
        ['({0})'.format(ZH_SYMBOLS)],
        element,
    )


# 1: wrong 「」.
def detect_e204(element):
    if not contains_chinese_characters(element.content):
        return False

    p = (
        r'('

        # ', " is handled by E201.
        # r"'"
        # r'|'
        # r'"'
        # r'|'

        # ‘’
        r'\u2018|\u2019'
        r'|'
        # “”
        r'\u201c|\u201d'

        r')'
    )
    return detect_by_patterns(
        [p],
        element,
    )


class SpecialWordHelper(object):

    WORD_PATTERN = {
        'App': r'app',
        'Android': r'android',
        'iOS': r'ios',
        'iPhone': r'iphone',
        'App Store': r'app\s?store',
        'WiFi': r'wi-*fi',
        'email': r'e-*mail',
        'P.S.': r'P\.*S\.*',
    }

    WORD_MAX_LENGTH = None
    SENTENCE_DELIMITER_TO_WORD = None

    @classmethod
    def init(cls):
        if cls.WORD_MAX_LENGTH and cls.SENTENCE_DELIMITER_TO_WORD:
            return

        delimiters = [
            '!', ';', '.', '?',
            '\uff01', '\uff1b', '\u3002', '\uff1f',
        ]

        cls.SENTENCE_DELIMITER_TO_WORD = defaultdict(list)
        for delimiter in delimiters:
            for word in cls.WORD_PATTERN:
                if delimiter not in word:
                    continue
                cls.SENTENCE_DELIMITER_TO_WORD[delimiter].append(word)

        cls.WORD_MAX_LENGTH = 0
        for word in cls.WORD_PATTERN:
            cls.WORD_MAX_LENGTH = max(cls.WORD_MAX_LENGTH, len(word))

    @classmethod
    def select_segment(cls, content, match):
        segment_begin = max(
            0,
            match.end() - SpecialWordHelper.WORD_MAX_LENGTH,
        )
        segment_end = min(
            len(content) - 1,
            match.start() + SpecialWordHelper.WORD_MAX_LENGTH,
        )
        return content[segment_begin:segment_end]

    @classmethod
    def delimiter_in_word(cls, content, match):
        delimiter = match.group(1)
        if delimiter not in cls.SENTENCE_DELIMITER_TO_WORD:
            return False

        segment = cls.select_segment(content, match)
        for word in cls.SENTENCE_DELIMITER_TO_WORD.get(delimiter, []):
            if segment.find(word) >= 0:
                return True
        return False


SpecialWordHelper.init()


# 1: wrong special word.
def detect_e301(element):

    for correct_form, pattern in SpecialWordHelper.WORD_PATTERN.items():

        # (?<!xxx) and (?!xxx) could match ^ and $.
        p = r'(?<![a-zA-Z])({0})(?![a-zA-Z])'.format(pattern)

        for m in re.finditer(
            p, element.content,
            flags=re.UNICODE | re.IGNORECASE,
        ):
            if m.group(0) != correct_form:
                yield m, correct_form


def process_errors_by_handler(error_codes, error_handler, element):

    for error_code in error_codes:
        detector = globals()['detect_{0}'.format(error_code.lower())]
        error_handler(error_code, element, detector(element))


def process_block_level_errors(error_handler, element):

    process_errors_by_handler(
        [
            'E101',
            'E102',
            'E103',
            'E104',

            'E203',
            'E205',
            'E206',
            'E207',

            'E301',
        ],
        error_handler,
        element,
    )


def split_text_element(element):

    # block_type should be split by newline first.
    SPLIT_BY_NEWLINES = [
        'list_block',
        'table',
    ]

    elements = []
    if element.block_type not in SPLIT_BY_NEWLINES:
        elements.append(element)
    else:
        content = element.content
        loc_begin = element.loc_begin

        if not content.strip('\n'):
            return []
        else:
            content = content.strip('\n')
            for line in content.split('\n'):
                elements.append(
                    TextElement(
                        'paragraph',
                        loc_begin, loc_begin,
                        line,
                    )
                )
                loc_begin += 1

    # split sentences.
    SENTENCE_DELIMITERS = (
        r'('
        r'\.{6}'
        r'|'
        r'!|;|\.|\?'
        r'|'
        r'\uff01|\uff1b|\u3002|\uff1f'
        r')'
    )

    OPEN_PARENTHESIS = set(
        '('
        '['
        "'"
        '"'
        '（'
        '「'
        '『'
    )
    CLOSE_PARENTHESIS = set(
        ')'
        ']'
        "'"
        '"'
        '）'
        '」'
        '』'
    )

    sentences = []

    for element in elements:
        content = element.content.strip('\n')

        loc_begin = element.loc_begin
        sbegin = 0

        for m in re.finditer(SENTENCE_DELIMITERS, content, flags=re.UNICODE):
            # ignore delimiter within special words.
            if SpecialWordHelper.delimiter_in_word(content, m):
                continue
            if delimiter_in_email(content, m):
                continue
            if delimiter_in_simple_uri(element.content, m):
                continue

            send = m.end()
            tailing_newlines = 0
            while send < len(content) and content[send] == '\n':
                send += 1
                tailing_newlines += 1

            sentence = content[sbegin:send]

            level = 0
            for c in sentence:
                if c in OPEN_PARENTHESIS:
                    level += 1
                elif c in CLOSE_PARENTHESIS:
                    level -= 1
            if level != 0:
                continue

            newlines = count_newlines(sentence)
            offset = count_offset(content[:sbegin])

            sentences.append(
                TextElement(
                    'paragraph',
                    loc_begin,
                    loc_begin + newlines - tailing_newlines,
                    sentence,
                    offset=offset,
                )
            )

            loc_begin += newlines
            sbegin = send

        if sbegin < len(content):
            sentences.append(
                TextElement(
                    'paragraph',
                    loc_begin, loc_begin,
                    content[sbegin:],
                    offset=count_offset(content[:sbegin]),
                )
            )
    return sentences


def process_sentence_level_errors(error_handler, element):

    process_errors_by_handler(
        [
            'E201',
            'E202',
            'E204',
        ],
        error_handler,
        element,
    )


def process_errors(error_handler, element):

    try_invoke(error_handler, 'before_block_level')
    process_block_level_errors(error_handler, element),
    try_invoke(error_handler, 'after_block_level')

    try_invoke(error_handler, 'before_sentence_level')
    for sentence in split_text_element(element):
        process_sentence_level_errors(error_handler, sentence)
    try_invoke(error_handler, 'after_sentence_level')



# -*- coding: utf-8 -*-
from __future__ import (
    division, absolute_import, print_function, unicode_literals,
)
from builtins import *                  # noqa
from future.builtins.disabled import *  # noqa

import re

from zhlint.error_correction import (
    ErrorCorrectionHandler,
    DiffOperation,
    INT_MAX,
)
from zhlint.error_detection import detect_e101  # noqa
from zhlint.error_detection import detect_e102  # noqa
from zhlint.error_detection import detect_e103  # noqa
from zhlint.error_detection import detect_e104  # noqa
from zhlint.error_detection import detect_e201  # noqa
from zhlint.error_detection import detect_e202  # noqa
from zhlint.error_detection import detect_e203  # noqa
from zhlint.error_detection import detect_e204  # noqa
from zhlint.error_detection import detect_e205  # noqa
from zhlint.error_detection import detect_e206  # noqa
from zhlint.error_detection import detect_e207  # noqa
from zhlint.error_detection import detect_e301  # noqa

from zhlint.utils import TextElement


SIMPLE_CONTENT = (
    'abc\n'
    '\n'
    'abc'
)
SIMPLE_ELEMENTS = (
    TextElement('paragraph', '1', '1', 'ac\n'),
    TextElement('paragraph', '3', '3', 'ac'),
    'EOF',
)


def test_generate_lcs_matrix():

    h = ErrorCorrectionHandler(SIMPLE_CONTENT, SIMPLE_ELEMENTS)

    l1 = [True] * (len(h.raw_lines[1]) + 1)
    l1[0] = l1[2] = False
    assert l1 == h.lcs_matrix[1]

    l3 = [True] * (len(h.raw_lines[3]) + 1)
    l3[0] = l3[2] = False
    assert l3 == h.lcs_matrix[3]

    l1 = [1, 3, 4]
    assert l1 == h.index_matrix[1]

    l3 = [1, 3]
    assert l3 == h.index_matrix[3]


def test_coordinate_query():

    h = ErrorCorrectionHandler(SIMPLE_CONTENT, SIMPLE_ELEMENTS)
    q = h.coordinate_query

    assert (1, 1) == q.query(0)
    assert (1, 4) == q.query(2)
    assert (3, 1) == q.query(3)
    assert (3, 3) == q.query(4)

    assert (3, 1) == q.query(0, base_loc=3)
    assert (3, 3) == q.query(1, base_loc=3)

    text = q.text()
    m = re.search(r'(ac)(\s+)(ac)', text)
    assert [
        [(1, 1), (1, 3)],
        [(1, 4)],
        [(3, 1), (3, 3)],
    ] == q.query_match(m)

    text = q.text(2, 3)
    m = re.search(r'(ac)', text)
    assert [
        [(3, 1), (3, 3)],
    ] == q.query_match(m, base_loc=2)

    # empty group.
    text = q.text()
    m = re.search(r'(a)()(c)', text)
    assert [
        [(1, 1)],
        [],
        [(1, 3)],
    ] == q.query_match(m)


def simple_init(error_code, text):
    te = TextElement('', 1, 1, text, offset=0)
    h = ErrorCorrectionHandler(
        text,
        (te, 'EOF'),
    )
    detector = globals()['detect_{0}'.format(error_code.lower())]
    h(error_code, te, detector(te))
    return h


def test_correct_e101():
    h = simple_init('E101', '中文english')
    assert [DiffOperation.insert(1, 3, val=' ')] == h.diffs

    h = simple_init('E101', 'english中文')
    assert [DiffOperation.insert(1, 8, val=' ')] == h.diffs

    h = simple_init('E101', '中文   english')
    assert [
        DiffOperation.delete(1, 3),
        DiffOperation.delete(1, 4),
        DiffOperation.delete(1, 5),
        DiffOperation.insert(1, 6, val=' '),
    ] == h.diffs

    h = simple_init('E101', 'english   中文')
    assert [
        DiffOperation.delete(1, 8),
        DiffOperation.delete(1, 9),
        DiffOperation.delete(1, 10),
        DiffOperation.insert(1, 11, val=' '),
    ] == h.diffs

    h = simple_init('E101', '中文\tenglish')
    assert [
        DiffOperation.delete(1, 3),
        DiffOperation.insert(1, 4, val=' '),
    ] == h.diffs

    h = simple_init('E101', 'english\t中文')
    assert [
        DiffOperation.delete(1, 8),
        DiffOperation.insert(1, 9, val=' '),
    ] == h.diffs


def test_correct_e102():
    h = simple_init('E102', '中文42')
    assert [DiffOperation.insert(1, 3, val=' ')] == h.diffs

    h = simple_init('E102', '42中文')
    assert [DiffOperation.insert(1, 3, val=' ')] == h.diffs

    h = simple_init('E102', '中文   42')
    assert [
        DiffOperation.delete(1, 3),
        DiffOperation.delete(1, 4),
        DiffOperation.delete(1, 5),
        DiffOperation.insert(1, 6, val=' '),
    ] == h.diffs

    h = simple_init('E102', '42   中文')
    assert [
        DiffOperation.delete(1, 3),
        DiffOperation.delete(1, 4),
        DiffOperation.delete(1, 5),
        DiffOperation.insert(1, 6, val=' '),
    ] == h.diffs

    h = simple_init('E102', '中文\t42')
    assert [
        DiffOperation.delete(1, 3),
        DiffOperation.insert(1, 4, val=' '),
    ] == h.diffs

    h = simple_init('E102', '42\t中文')
    assert [
        DiffOperation.delete(1, 3),
        DiffOperation.insert(1, 4, val=' '),
    ] == h.diffs


def test_correct_e103():
    h = simple_init('E103', '42μ')
    assert [DiffOperation.insert(1, 3, val=' ')] == h.diffs

    h = simple_init('E103', '42  μ')
    assert [
        DiffOperation.delete(1, 3),
        DiffOperation.delete(1, 4),
        DiffOperation.insert(1, 5, val=' '),
    ] == h.diffs


def test_correct_e104():
    h = simple_init('E104', '(42）')
    assert [
        DiffOperation.replace(1, 4, val=')'),
    ] == h.diffs

    h = simple_init('E104', '（42）')
    assert [
        DiffOperation.replace(1, 1, val='('),
        DiffOperation.replace(1, 4, val=')'),
    ] == h.diffs

    h = simple_init('E104', '42(42)')
    assert [
        DiffOperation.insert(1, 3, val=' '),
    ] == h.diffs

    h = simple_init('E104', '42  (42)')
    assert [
        DiffOperation.delete(1, 3),
        DiffOperation.delete(1, 4),
        DiffOperation.insert(1, 5, val=' '),
    ] == h.diffs

    h = simple_init('E104', '  (42)')
    assert [
        DiffOperation.delete(1, 1),
        DiffOperation.delete(1, 2),
    ] == h.diffs


def test_correct_e201():
    h = simple_init('E201', '有中文,错误.')
    assert [
        DiffOperation.replace(1, 4, val='，'),
        DiffOperation.replace(1, 7, val='。'),
    ] == h.diffs

    h = simple_init('E201', '有中文, 错误.')
    assert [
        DiffOperation.replace(1, 4, val='，'),
        DiffOperation.delete(1, 5),
        DiffOperation.replace(1, 8, val='。'),
    ] == h.diffs

    h = simple_init('E201', '有中文"错误"')
    assert [
        DiffOperation.replace(1, 4, val='「'),
        DiffOperation.replace(1, 7, val='」'),
    ] == h.diffs

    h = simple_init('E201', '有中文(错误)')
    assert [
        DiffOperation.replace(1, 4, val='（'),
        DiffOperation.replace(1, 7, val='）'),
    ] == h.diffs

    h = simple_init('E201', ' 步的重定向 transitions $$.')
    assert [
        DiffOperation.replace(1, 22, val='。'),
    ] == h.diffs


def test_correct_e202():
    h = simple_init('E202', 'english，test。')
    assert [
        DiffOperation.replace(1, 8, val=', '),
        DiffOperation.replace(1, 13, val='. '),
    ] == h.diffs


def test_correct_e203():
    h = simple_init('E203', '中文， 测试')
    assert [
        DiffOperation.delete(1, 4),
    ] == h.diffs

    h = simple_init('E203', '中文 ，测试')
    assert [
        DiffOperation.delete(1, 3),
    ] == h.diffs


def test_correct_e204():
    h = simple_init('E204', '中文“测试”')
    assert [
        DiffOperation.replace(1, 3, '「'),
        DiffOperation.replace(1, 6, '」'),
    ] == h.diffs

    h = simple_init('E204', '中文‘测试’')
    assert [
        DiffOperation.replace(1, 3, '「'),
        DiffOperation.replace(1, 6, '」'),
    ] == h.diffs


def test_correct_e205():

    h = simple_init('E205', '中文...')
    assert [
        DiffOperation.delete(1, 3),
        DiffOperation.delete(1, 4),
        DiffOperation.delete(1, 5),
        DiffOperation.insert(INT_MAX, INT_MAX, '......'),
    ] == h.diffs

    h = simple_init('E205', '中文.......')
    assert [
        DiffOperation.delete(1, 3),
        DiffOperation.delete(1, 4),
        DiffOperation.delete(1, 5),
        DiffOperation.delete(1, 6),
        DiffOperation.delete(1, 7),
        DiffOperation.delete(1, 8),
        DiffOperation.delete(1, 9),
        DiffOperation.insert(INT_MAX, INT_MAX, '......'),
    ] == h.diffs

    h = simple_init('E205', '中文。。')
    assert [
        DiffOperation.delete(1, 3),
        DiffOperation.delete(1, 4),
        DiffOperation.insert(INT_MAX, INT_MAX, '......'),
    ] == h.diffs


def test_correct_e206():
    h = simple_init('E206', 'this is sparta!!!')
    assert [
        DiffOperation.delete(1, 15),
        DiffOperation.delete(1, 16),
        DiffOperation.delete(1, 17),
        DiffOperation.insert(INT_MAX, INT_MAX, '!'),
    ] == h.diffs

    h = simple_init('E206', 'this is sparta???')
    assert [
        DiffOperation.delete(1, 15),
        DiffOperation.delete(1, 16),
        DiffOperation.delete(1, 17),
        DiffOperation.insert(INT_MAX, INT_MAX, '?'),
    ] == h.diffs


def test_correct_e207():
    h = simple_init('E207', '讨厌啦~')
    assert [
        DiffOperation.delete(1, 4),
    ] == h.diffs

    h = simple_init('E207', '讨厌啦~~~~')
    assert [
        DiffOperation.delete(1, 4),
        DiffOperation.delete(1, 5),
        DiffOperation.delete(1, 6),
        DiffOperation.delete(1, 7),
    ] == h.diffs


def test_correct_e301():
    h = simple_init('E301', 'APP')
    assert [
        DiffOperation.delete(1, 1),
        DiffOperation.delete(1, 2),
        DiffOperation.delete(1, 3),
        DiffOperation.insert(INT_MAX, INT_MAX, val='App'),
    ] == h.diffs



# -*- coding: utf-8 -*-
from __future__ import (
    division, absolute_import, print_function, unicode_literals,
)
from builtins import *                  # noqa
from future.builtins.disabled import *  # noqa

import re
import itertools
import collections as abc

import click

from zhlint.utils import count_newlines


ERROR_DISPLAY = {
    # space.
    'E101': '英文与非标点的中文之间需要有一个空格',
    'E102': '数字与非标点的中文之间需要有一个空格',
    'E103': '除了「％」、「℃」、以及倍数单位（如 2x、3n）之外，'
            '其余数字与单位之间需要加空格',
    'E104': '书写时括号中全为数字，则括号用半角括号且首括号前要空一格',

    # punctuation.
    'E201': '只有中文或中英文混排中，一律使用中文全角标点',
    'E202': '如果出现整句英文，则在这句英文中使用英文、半角标点',
    'E203': '中文标点与其他字符间一律不加空格',
    'E204': '中文文案中使用中文引号「」和『』，其中「」为外层引号',
    'E205': '省略号请使用「……」标准用法',
    'E206': '感叹号请使用「！」标准用法',
    'E207': '请勿在文章内使用「~」',

    # terminology.
    'E301': '常用名词错误',
}


def split_bar(symbol, offset, *texts):
    length = offset
    for text in texts:
        for c in text:
            if 0 <= ord(c) <= 127:
                length += 1
            else:
                length += 2
    return click.style(symbol * length)


def get_loc(element, i, j):
    begin = element.loc_begin
    begin += count_newlines(element.content[:i])
    end = begin + count_newlines(element.content[i:j])
    return begin, end


def generate_detected_text(element, i, j):

    def inc(char, count):
        if 0 <= ord(char) <= 127:
            # for ASCII, increase by 1.
            return count + 1
        else:
            # otherwise, increase by 2.
            return count + 2

    OFFSET = 20
    content = element.content

    # expand to left.
    ri = i
    c = 0
    while ri >= 0 and c < OFFSET and content[ri] != '\n':
        ri -= 1
        c = inc(content[ri], c)
    ri = min(ri + 1, i)

    # expand to right.
    rj = j
    c = 0
    while rj < len(content) and c < OFFSET and content[ri] != '\n':
        rj += 1
        c = inc(content[ri], c)
    rj = max(rj - 1, j)

    text_line = content[ri:rj]
    i = i - ri
    j = j - ri

    for c in ['\t', '\n']:
        for m in re.finditer(c, text_line):
            ci = m.start()
            if ci < i:
                i += 5
                j += 5
            elif ci < j:
                j += 5

        text_line = text_line.replace(
            c,
            ' [{0}] '.format(repr(c)[2:-1]),
        )

    mark_line = []
    for c in text_line[:i]:
        if 0 <= ord(c) <= 127:
            mark_line.append(' ')
        else:
            mark_line.append('\u3000')
    for c in text_line[i:j]:
        if 0 <= ord(c) <= 127:
            mark_line.append('-')
        else:
            mark_line.append('\uff0d')
    mark_line = ''.join(mark_line)

    return text_line, mark_line


class ErrorDisplayHandler(object):

    def __init__(self):
        self.detected_error = False

    def header_styles(self, code):
        return [
            split_bar('=', 2, code, ERROR_DISPLAY[code]),
            click.style('\n'),
            click.style(code, fg='green', bold=True),
            click.style(': '),
            click.style(ERROR_DISPLAY[code]),
            click.style('\n'),
            split_bar('=', 2, code, ERROR_DISPLAY[code]),
            click.style('\n'),
        ]

    def body_styles(self, element, matches):

        loc_detected = []
        for m in matches:
            # support context.
            if isinstance(m, abc.Sized) and len(m) == 2:
                m = m[0]

            loc_detected.append(
                (m.start(), m.end()),
            )

        styles = []
        for i, j in sorted(loc_detected):
            loc = get_loc(element, i, j)
            text_line, mark_line = generate_detected_text(element, i, j)

            if loc[0] == loc[1]:
                loc_text = str(loc[0])
            else:
                loc_text = '{0}-{1}'.format(*map(str, loc))

            styles.extend([
                click.style('LINE', fg='green', bold=True),
                click.style(': ', fg='white'),
                click.style(loc_text, fg='red'),
                click.style('\n'),
                click.style(text_line, fg='yellow'),
                click.style('\n'),
                click.style(mark_line, fg='red', bold=True),
                click.style('\n'),
                split_bar('.', 0, text_line),
                click.style('\n'),
            ])
        return styles

    def __call__(self, error_code, element, matches):
        if not matches:
            return
        matches = list(matches)

        if not matches:
            return
        else:
            self.detected_error = True

        for style in itertools.chain(
            self.header_styles(error_code),
            self.body_styles(element, matches)
        ):
            click.echo(style, nl=False)
        click.echo('\n', nl=False)



# -*- coding: utf-8 -*-
from __future__ import (
    division, absolute_import, print_function, unicode_literals,
)
from builtins import *                  # noqa
from future.builtins.disabled import *  # noqa

import os.path

from mistune import preprocessing

from zhlint.preprocessor import transform


DATA = os.path.join(
    os.path.dirname(__file__),
    'data',
)


def load_test_md(name):
    return preprocessing(
        open(os.path.join(DATA, name), encoding='utf-8').read(),
    )


def eof(element):
    assert 'EOF' == element.content


def test_latex_inline():
    elements = transform(load_test_md('latex_inline.md'))
    assert 'a line with $$ words.\n\n' == elements[0].content
    assert 'a line with \\(\\) words.\n\n' == elements[1].content
    assert '会使 $$ 加入到 $$ 中\n' == elements[2].content
    eof(elements[3])


def test_latex_block():
    elements = transform(load_test_md('latex_block.md'))
    assert 'block 1\n\n\n\n' == elements[0].content
    assert 'block 2\n\n\n\n\n\n' == elements[1].content
    eof(elements[2])


def test_yaml_header():
    elements = transform(load_test_md('yaml_header.md'))
    assert 'test line.\n' == elements[0].content
    eof(elements[1])


def assert_loc(begin, end, element):
    assert begin == int(element.loc_begin)
    assert end == int(element.loc_end)


def test_loc():
    elements = transform(load_test_md('loc.md'))

    assert_loc(5, 8, elements[0])
    assert_loc(9, 15, elements[1])
    assert_loc(16, 17, elements[2])
    eof(elements[3])


def test_ref_loc():
    elements = transform(load_test_md('ref.md'))

    assert 5 == len(elements)
    assert_loc(2, 3, elements[0])
    assert_loc(4, 6, elements[1])
    assert_loc(7, 8, elements[2])
    assert_loc(9, 12, elements[3])
    eof(elements[4])


def test_list_block():
    elements = transform(load_test_md('list_block.md'))

    assert 3 == len(elements)
    assert_loc(2, 7, elements[0])
    assert_loc(8, 8, elements[1])

    eof(elements[2])


def test_table():
    elements = transform(load_test_md('table.md'))
    assert 1 == len(elements)
    eof(elements[0])


def test_link():
    elements = transform(load_test_md('link.md'))
    assert 3 == len(elements)
    assert_loc(3, 11, elements[0])
    assert_loc(12, 12, elements[1])
    eof(elements[2])

    lines = elements[0].content.split('\n')
    expected = [
        '',
        'this',
        'is',
        'a',
        'long',
        'link',
        'a',
        'this is test',
        # this is a test\n(1)\n(2)
        '',
        '',
    ]
    assert expected == lines


def test_link_ending():
    elements = transform(load_test_md('link_ending.md'))
    assert 2 == len(elements)
    eof(elements[1])


def test_newline():
    elements = transform(load_test_md('newline.md'))
    assert 2 == len(elements)
    assert_loc(1, 5, elements[0])
    eof(elements[1])


def test_nested_list():
    elements = transform(load_test_md('nested_list.md'))
    assert 7 == len(elements)

    assert 'a 1.\nb 1.\nb 2.\na 2.\nb 3.\nb 4.\n\n' == elements[0].content
    assert_loc(1, 7, elements[0])

    assert_loc(8, 9, elements[1])

    assert 'line\n\nref line.\n\n' == elements[2].content
    assert_loc(10, 13, elements[2])

    assert_loc(14, 15, elements[3])

    assert 'line\n\n\n\n\n\n' == elements[4].content
    assert_loc(16, 21, elements[4])

    assert_loc(22, 22, elements[5])

    eof(elements[6])


def test_seperated_list():
    elements = transform(load_test_md('seperated_list.md'))
    assert 3 == len(elements)
    assert 'line 1\n\nline 2\n\n' == elements[0].content
    assert_loc(1, 4, elements[0])
    assert_loc(5, 5, elements[1])
    eof(elements[2])


def test_tailing_list():
    elements = transform(load_test_md('tailing_list.md'))
    assert 2 == len(elements)
    assert_loc(1, 9, elements[0])
    eof(elements[1])



# -*- coding: utf-8 -*-
from __future__ import (
    division, absolute_import, print_function, unicode_literals,
)


NAME = 'zhlint'
VERSION = '0.2.1'
AUTHORS = [
    'huntzhan',
]
EMAILS = [
    'programmer.zhx@gmail.com',
]
LICENSE = ''
URL = ''
DESCRIPTION = 'automated checker of chinese document.'



# -*- coding: utf-8 -*-
from __future__ import (
    division, absolute_import, print_function, unicode_literals,
)
from builtins import *                  # noqa
from future.builtins.disabled import *  # noqa

from itertools import chain
import functools

from zhlint.lcs import lcs_marks
from zhlint.utils import text2lines, safelen


INT_MAX = 1 << 31 - 1


@functools.total_ordering
class DiffOperation(object):

    INSERT = 0
    DELETE = 1
    REPLACE = 2

    @staticmethod
    def insert(*args, **kwargs):
        return DiffOperation(DiffOperation.INSERT, *args, **kwargs)

    @staticmethod
    def delete(*args, **kwargs):
        return DiffOperation(DiffOperation.DELETE, *args, **kwargs)

    @staticmethod
    def replace(*args, **kwargs):
        return DiffOperation(DiffOperation.REPLACE, *args, **kwargs)

    def __init__(self, tag, row, col, val=None, parent=None):
        self.tag = tag
        self.row = row
        self.col = col
        self.val = val
        self.parent = parent

        self._attrs = (self.row, self.col, self.tag, self.parent, self.val)

    def __eq__(self, other):
        return self._attrs == other._attrs

    def __lt__(self, other):
        return self._attrs < other._attrs

    def __str__(self):
        TAGS = ['INSERT', 'DELETE', 'REPLACE']
        attrs = list(self._attrs)
        attrs[2] = TAGS[attrs[2]]
        return '<{0}>'.format(', '.join(map(str, attrs)))

    def __repr__(self):
        return str(self).encode('utf-8')


def add_diff_operations(attr, coordinates, diffs, val=None):
    func = getattr(DiffOperation, attr)
    ops = [func(x, y, val=val) for x, y in coordinates]
    diffs.extend(ops)


def delete_group(coordinates, diffs):
    add_diff_operations('delete', coordinates, diffs)


def insert_before(coordinate, diffs, val):
    add_diff_operations('insert', [coordinate], diffs, val=val)


def replace_with(coordinate, diffs, val):
    add_diff_operations('replace', [coordinate], diffs, val=val)


def correct_single_space_problem(element, match, handler):

    coordinates = handler.coordinate_query.query_match(
        match, base_loc=element.loc_begin,
    )
    a, whitespaces, b = match.groups()

    if whitespaces:
        # delete all whitespaces.
        delete_group(coordinates[1], handler.diffs)

    # no whitespaces between a and b.
    # insert space between a and b.
    insert_before(coordinates[2][0], handler.diffs, val=' ')


def correct_e101(element, match, handler):
    correct_single_space_problem(element, match, handler)


def correct_e102(element, match, handler):
    correct_single_space_problem(element, match, handler)


def correct_e103(element, match, handler):
    correct_single_space_problem(element, match, handler)


def correct_e104(element, match, handler):
    match, tag = match
    coordinates = handler.coordinate_query.query_match(
        match, base_loc=element.loc_begin,
    )
    whitespaces, lparenthesis, digits, rparenthesis = match.groups()

    # deal with spaces.
    if tag != 0:
        delete_group(coordinates[0], handler.diffs)
        if tag == 1:
            insert_before(coordinates[1][0], handler.diffs, val=' ')

    if lparenthesis != '(':
        replace_with(coordinates[1][0], handler.diffs, val='(')
    if rparenthesis != ')':
        replace_with(coordinates[3][0], handler.diffs, val=')')


class PunctuationConverter(object):

    EN2ZH = {
        # positive.
        '!': '！',
        '$': '＄',
        '(': '（',
        ')': '）',
        ',': '，',
        '.': '。',
        ':': '：',
        ';': '；',
        '<': '《',
        '>': '》',
        '?': '？',
        '\\': '、',
        '/': '、',
        '_': '＿',

        # best effort.
        '[': '「',
        ']': '」',
        '{': '『',
        '}': '』',

        # not sure.
        # https://en.wikipedia.org/wiki/Caret
        '^': '\u2038',
    }

    ZH2EN = {
        # positive.
        '！': '!',
        '＄': '$',
        '（': '(',
        '）': ')',
        '，': ',',
        '。': '.',
        '：': ':',
        '；': ';',
        '《': '<',
        '》': '>',
        '？': '?',
        '、': ',',
        '＿': '_',

        '‘': "'",
        '’': "'",
        '“': '"',
        '”': '"',

        # best effort.
        '「': '"',
        '」': '"',
        '『': '"',
        '』': '"',

        # not sure.
        # https://en.wikipedia.org/wiki/Caret
        '\u2038': '^',
    }


def correct_e201(element, match, handler):

    def guess_open_or_close(text, punctuation):
        return sum(map(lambda c: 1 if c == punctuation else 0, text)) % 2 == 0

    coordinates = handler.coordinate_query.query_match(
        match, base_loc=element.loc_begin,
        offset=element.offset,
    )
    punctuation, whitespaces = match.groups()

    if punctuation in PunctuationConverter.EN2ZH:
        replace_with(
            coordinates[0][0], handler.diffs,
            val=PunctuationConverter.EN2ZH[punctuation],
        )
    else:
        # ' and "
        if guess_open_or_close(element.content[:match.start()], punctuation):
            val = '「'
        else:
            val = '」'
        replace_with(
            coordinates[0][0], handler.diffs, val=val,
        )

    if whitespaces:
        delete_group(coordinates[1], handler.diffs)


def correct_e202(element, match, handler):

    coordinates = handler.coordinate_query.query_match(
        match, base_loc=element.loc_begin,
        offset=element.offset,
    )
    punctuation = match.group(1)

    if punctuation in PunctuationConverter.ZH2EN:
        replace_with(
            coordinates[0][0], handler.diffs,
            # because of E203, a space should be added to en punctuation.
            val=PunctuationConverter.ZH2EN[punctuation] + ' ',
        )


def correct_e203(element, match, handler):
    coordinates = handler.coordinate_query.query_match(
        match, base_loc=element.loc_begin,
    )
    a, whitespaces, b = match.groups()

    if whitespaces:
        delete_group(coordinates[1], handler.diffs)


def correct_e204(element, match, handler):

    ZH_QUOTE_MAPPING = {
        '\u2018': '「',
        '\u201c': '「',
        '\u2019': '」',
        '\u201d': '」',
    }

    coordinates = handler.coordinate_query.query_match(
        match, base_loc=element.loc_begin,
    )
    quote = match.group(1)
    replace_with(
        coordinates[0][0], handler.diffs,
        ZH_QUOTE_MAPPING[quote],
    )


def correct_e205(element, match, handler):

    coordinates = handler.coordinate_query.query_match(
        match, base_loc=element.loc_begin,
        add_right_boundary=True,
    )
    delete_group(coordinates[0], handler.diffs)
    insert_before(coordinates[1][0], handler.diffs, val='......')


def correct_e206(element, match, handler):

    coordinates = handler.coordinate_query.query_match(
        match, base_loc=element.loc_begin,
        add_right_boundary=True,
    )
    mark = match.group(1)[0]
    delete_group(coordinates[0], handler.diffs)
    insert_before(coordinates[1][0], handler.diffs, val=mark)


def correct_e207(element, match, handler):

    coordinates = handler.coordinate_query.query_match(
        match, base_loc=element.loc_begin,
    )
    delete_group(coordinates[0], handler.diffs)


def correct_e301(element, match, handler):
    match, correct_form = match

    coordinates = handler.coordinate_query.query_match(
        match, base_loc=element.loc_begin,
        add_right_boundary=True,
    )
    delete_group(coordinates[0], handler.diffs)
    insert_before(coordinates[1][0], handler.diffs, val=correct_form)


class CoordinateQuery(object):

    def __init__(self, index_matrix, lcs_matrix, parsed_lines):
        # line_size[i]: the number of characters in line i.
        line_size = [
            len(line)
            for line in chain('', map(lambda x: x or '', parsed_lines))
        ]

        # line_acc[i]: the number of characters from line 1 to line i.
        self.line_acc = [0] * len(line_size)
        for i in range(1, len(self.line_acc)):
            self.line_acc[i] = self.line_acc[i - 1] + line_size[i]

        self.parsed_lines = parsed_lines
        self.index_matrix = index_matrix
        self.lcs_matrix = lcs_matrix

    # mainly for testing.
    # return: text from line begin to line end, including end.
    def text(self, begin=1, end=None):
        end = end or len(self.parsed_lines) - 1
        end += 1
        return ''.join(map(lambda x: x or '', self.parsed_lines[begin:end]))

    # input:
    # offset: 0-based.
    # base_loc: 1-based.
    #
    # return: coordinate of original matrix.
    # row: 1-based.
    # col: 1-based.
    def query(self, parsed_offset, base_loc=1):
        base_acc = self.line_acc[base_loc - 1]
        lo = base_loc
        hi = len(self.line_acc) - 1
        while lo < hi:
            mid = lo + (hi - lo) // 2
            if parsed_offset + 1 <= (self.line_acc[mid] - base_acc):
                hi = mid
            else:
                lo = mid + 1

        # 1-based row index.
        row = lo
        # 0-based.
        col = parsed_offset - self.line_acc[lo - 1] + base_acc
        # corner case: reach EOF.
        if (
            row == len(self.index_matrix) - 1 and
            col == len(self.index_matrix[row])
        ):
            return None, None

        # 1-based col index.
        col = self.index_matrix[row][col]
        return row, col

    # return:
    # [((x1, y1), (x2, y2), ...)...]
    def query_match(self, match, offset=0, base_loc=1,
                    add_right_boundary=False):
        match_offset = match.start()
        group_sizes = [len(g) for g in match.groups([])]
        if add_right_boundary:
            group_sizes.append(1)

        coordinates = []
        acc = 0
        for size in group_sizes:
            if size == 0:
                coordinates.append([])
                continue

            start = self.query(match_offset + acc + offset, base_loc)
            end = self.query(match_offset + acc + offset + size - 1, base_loc)

            group_coordinates = []
            row, col = start
            # reach EOF.
            if row is None and col is None:
                coordinates.append(
                    [(INT_MAX, INT_MAX)]
                )
                break

            while (row, col) <= end:
                if self.lcs_matrix[row][col]:
                    group_coordinates.append(
                        (row, col),
                    )
                col += 1
                if col == len(self.lcs_matrix[row]):
                    row += 1
                    col = 0

            coordinates.append(group_coordinates)
            acc += size

        return coordinates


class DiffOperationExecutor(object):

    def __init__(self, diffs, raw_lines):
        self.diffs = self.postprocess_diffs(diffs[:])
        self.raw_lines = self.postprocess_raw_lines(raw_lines)

    def postprocess_raw_lines(self, raw_lines):
        # convert to 1-based.
        return [list(chain([None], line or [])) for line in raw_lines]

    def postprocess_diffs(self, diffs):
        # 1. sort and merge.
        diffs = sorted(diffs)
        merged_diffs = []
        i = 0
        n = len(diffs)
        while i < n:
            j = i + 1
            while j < n and diffs[i] == diffs[j]:
                j += 1
            merged_diffs.append(diffs[i])
            i = j
        diffs = merged_diffs

        # 2. check conflict.
        i = 0
        n = len(diffs)
        while i < n:
            op_i = diffs[i]
            j = i + 1
            while j < n:
                op_j = diffs[j]
                if op_i.row == op_j.row and op_i.col == op_j.col:
                    j += 1
                else:
                    break

            op_group = diffs[i:j]
            op_tags = [op.tag for op in op_group]

            if len(op_group) >= 2:
                if (
                    len(op_group) > 2 or
                    op_tags[0] != DiffOperation.INSERT or
                    op_tags[1] == DiffOperation.INSERT
                ):
                    raise RuntimeError('Detect Conflicts!')

            # move to next.
            i = j

        return diffs

    def _check_axis(self, row, col, di, diffs):
        if di == len(diffs):
            return True
        else:
            op = diffs[di]
            return (row, col) < (op.row, op.col)

    def _apply_diff_operation(self, col, op, content):
        if op.tag == DiffOperation.INSERT:
            content.append(op.val)
        elif op.tag == DiffOperation.DELETE:
            col += 1
        elif op.tag == DiffOperation.REPLACE:
            content.append(op.val)
            col += 1
        return col

    def apply_diff_operations(self):
        m = len(self.raw_lines) - 1
        di = 0
        content = []

        for row in range(1, m + 1):
            n = len(self.raw_lines[row]) - 1
            col = 1
            while col <= n:
                if self._check_axis(row, col, di, self.diffs):
                    # normal case.
                    content.append(self.raw_lines[row][col])
                    col += 1
                else:
                    # make changes.
                    col = self._apply_diff_operation(
                        col,
                        self.diffs[di],
                        content,
                    )
                    di += 1

        # EOF.
        if di == len(self.diffs) - 1:
            content.append(self.diffs[di].val)

        return ''.join(content)


class ErrorCorrectionHandler(object):

    # build block-level LCS.
    def __init__(self, file_content, elements):
        elements = elements[:-1]
        # init self.lcs_matrix and self.raw_lines
        self.generate_lcs_matrix(file_content, elements)
        # init self.parsed_lines
        self.generate_parsed_lines(elements)
        # init self.index_matrix
        self.generate_index_matrix()

        # store diff operations.
        self.diffs = []

        # init CoordinateQuery.
        self.coordinate_query = CoordinateQuery(
            self.index_matrix, self.lcs_matrix, self.parsed_lines,
        )

    @property
    def max_linenum(self):
        return len(self.raw_lines)

    def raw_line(self, i):
        return self.raw_lines[i], safelen(self.raw_lines[i])

    def parsed_line(self, i):
        return self.parsed_lines[i], safelen(self.parsed_lines[i])

    def generate_lcs_matrix(self, file_content, elements):
        # one-based.
        lines = text2lines(file_content)
        lines.insert(0, [])

        self.lcs_matrix = [[False] * (len(line) + 1) for line in lines]

        self.raw_lines = lines
        self.raw_lines[0] = None

        for element in elements:
            lbegin = element.loc_begin
            lend = element.loc_end

            x = ''.join(chain(*lines[lbegin:lend + 1]))
            y = element.content

            marks = lcs_marks(x, y)

            row = lbegin
            col = 1
            mi = 0
            while row <= lend:
                self.lcs_matrix[row][col] = marks[mi]
                mi += 1
                if col == len(self.lcs_matrix[row]) - 1:
                    row += 1
                    col = 1
                else:
                    col += 1

    def generate_parsed_lines(self, elements):
        self.parsed_lines = [None] * self.max_linenum
        for element in elements:
            lbegin = element.loc_begin
            for i, line in enumerate(text2lines(element.content)):
                self.parsed_lines[lbegin + i] = line

    def generate_index_matrix(self):
        self.index_matrix = [None] * self.max_linenum
        for row in range(1, self.max_linenum):
            parsed_line, ln = self.parsed_line(row)
            index_mapping = [None] * ln

            rawcol = 1
            for col in range(ln):
                while not self.lcs_matrix[row][rawcol]:
                    rawcol += 1
                index_mapping[col] = rawcol
                rawcol += 1

            self.index_matrix[row] = index_mapping

    def __call__(self, error_code, element, matches):
        if not matches:
            return

        corrector = globals()['correct_{0}'.format(error_code.lower())]
        for match in matches:
            corrector(element, match, self)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import (
    division, absolute_import, print_function, unicode_literals,
)

import codecs
import imp
import os
from setuptools import setup, find_packages

# load metadata.
metadata = imp.load_source(
    'metadata',
    os.path.join('zhlint', 'metadata.py'),
)


def read(fname):
    file_path = os.path.join(os.path.dirname(__file__), fname)
    return codecs.open(file_path, encoding='utf-8').read()


def load_requirements(fname):
    # only handles a simple subset of requirements format.
    pkgs = []
    for line in read(fname).split(os.linesep):
        if not line or line.startswith('#'):
            continue
        pkgs.append(line)
    return pkgs


setup(
    # informations.
    name=metadata.NAME,
    version=metadata.VERSION,
    author=metadata.AUTHORS[0],
    author_email=metadata.EMAILS[0],
    maintainer=metadata.AUTHORS[0],
    maintainer_email=metadata.EMAILS[0],
    license=metadata.LICENSE,
    url=metadata.URL,
    description=metadata.DESCRIPTION,
    long_description=read('README.md'),
    # https://pypi.python.org/pypi?:action=list_classifiers
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Development Status :: 1 - Planning',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
    # critical configurations.
    packages=find_packages(),
    install_requires=load_requirements('requirements.txt'),
    entry_points={
        'console_scripts': [
            'zhlint = zhlint.main:entry_point'
        ],
    },
)



# -*- coding: utf-8 -*-
from __future__ import (
    division, absolute_import, print_function, unicode_literals,
)
from builtins import *                  # noqa
from future.builtins.disabled import *  # noqa

import os.path

from zhlint.error_detection import (
    process_errors,
)
from zhlint.preprocessor import transform
from zhlint.error_correction import (
    ErrorCorrectionHandler,
    DiffOperation,
    DiffOperationExecutor,
)


DATA = os.path.join(
    os.path.dirname(__file__),
    'data',
)


def load_test_md(name):
    return open(os.path.join(DATA, name), encoding='utf-8').read()


def load_and_check(name):
    content = load_test_md(name)
    elements = transform(content)
    h = ErrorCorrectionHandler(content, elements)
    process_errors(h, elements[0])
    executor = DiffOperationExecutor(h.diffs, h.raw_lines)
    return executor


def test_simple():
    executor = load_and_check('correction_simple.md')
    assert [
        DiffOperation.replace(1, 3, val='，'),
        DiffOperation.delete(1, 4),
        DiffOperation.replace(1, 7, val='。'),
    ] == executor.diffs

    assert '中文，标点。\n' == executor.apply_diff_operations()


def test_latex():
    executor = load_and_check('correction_latex.md')
    assert [
        DiffOperation.replace(1, 40, val='。'),
    ] == executor.diffs

    assert (
        '第 (5) 步的重定向 transitions $ChildrenTrans$。\n' ==
        executor.apply_diff_operations()
    )



# -*- coding: utf-8 -*-
from __future__ import (
    division, absolute_import, print_function, unicode_literals,
)
from builtins import *                  # noqa
from future.builtins.disabled import *  # noqa

from zhlint.lcs import lcs_marks


def test_simple():
    x = 'abbc'
    y = 'abc'

    assert [True, True, False, True] == lcs_marks(x, y)


def test_header():
    x = '### abcd'
    y = 'abc'

    expected = [False] * len(x)
    expected[4:7] = [True] * 3
    assert expected == lcs_marks(x, y)


def test_newline():
    x = 'a\n\nb'
    y = 'a\nb'

    expected = [True, True, False, True]
    assert expected == lcs_marks(x, y)


def test_latex():
    x = '3. 第 (5) 步的重定向 transitions $ChildrenTrans$.\n'
    y = '第 (5) 步的重定向 transitions $$.\n'
    expected = [True] * len(x)
    expected[0:3] = [False] * 3
    expected[28:41] = [False] * (41 - 28)
    assert expected == lcs_marks(x, y)
